unit SpinHex;

interface

uses
  Classes, Controls, Dialogs, Math, Messages, Spin, StdCtrls, StrUtils,
  SysUtils, Windows;

type
  TEasyHex = class(TCustomEdit)
  private
    FValueMax: integer;   // 
    FValue: integer;      // 
    FValueMin: integer;   // 
    FIsSystemHex: Boolean;  //  16-  ?
    function GetValueMax: integer;
    procedure SetValueMax(v: integer);
    function GetValue: integer;
    procedure SetValue(v: integer);
    function GetValueMin: integer;
    procedure SetValueMin(v: integer);
    function GetIsSystemHex: Boolean;
    procedure SetIsSystemHex(v: Boolean);
    procedure ValueToText;
    function DifferenceValue(v: integer): integer;

  protected
    procedure KeyDown(var key: Word; shift: TShiftState); override;
    procedure KeyPress(var key: Char); override;
    procedure Change; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  published
    property ValueMax: integer read GetValueMax write SetValueMax;
    property Value: integer read GetValue write SetValue;
    property ValueMin: integer read GetValueMin write SetValueMin;
    property IsSystemHex: Boolean read GetIsSystemHex write SetIsSystemHex;
    property Anchors;
    property AutoSelect;
    property AutoSize;
    property Color;
    property Enabled;
    property Font;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;



  TSpinHex = class(TEasyHex)
  private
    FButton: TSpinButton;
    procedure SetEditRect;                                   //   
    procedure WMSize(var Message: TWMSize); message WM_SIZE; // 
    function GetMinHeight: Integer;                          // ?
    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;

  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure UpClick (Sender: TObject); virtual;
    procedure DownClick (Sender: TObject); virtual;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  published

  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Standard', [TEasyHex,TSpinHex]);
end;



{   TEasyHex }

constructor TEasyHex.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Text := '0';    
end;

destructor TEasyHex.Destroy;
begin
  inherited Destroy;
end;



function TEasyHex.GetValueMax: Integer;
begin
  result := FValueMax;
end;

procedure TEasyHex.SetValueMax (v: Integer);
begin
  If v<FValueMin then v:=FValueMin;
  If v<FValue then begin
                   FValue := v;
                   ValueToText;
                   end;
  FValueMax := v;
end;



function TEasyHex.GetValue: Integer;
begin
  If FValue<FValueMin then begin
                           FValue := FValueMin;
                           ValueToText;
                           end;
  If Text='' then ValueToText;
  result := FValue;
end;

procedure TEasyHex.SetValue (v: Integer);
begin
  FValue := EnsureRange(v,FValueMin,FValueMax);
  ValueToText;
end;



function TEasyHex.GetValueMin: Integer;
begin
  result := FValueMin;
end;

procedure TEasyHex.SetValueMin (v: Integer);
begin
  v := EnsureRange(v,0,FValueMax);
  If FValue<v then begin
                   FValue := v;
                   ValueToText;
                   end;
  FValueMin := v;
end;



function TEasyHex.GetIsSystemHex: Boolean;
begin
  result := FIsSystemHex;
end;

procedure TEasyHex.SetIsSystemHex (v: Boolean);
begin
  If FIsSystemHex<>v
    then begin
         FIsSystemHex := v;
         ValueToText;
         end;
end;



procedure TEasyHex.ValueToText;
var i: Integer;
begin
  try                                //     
     i := SelStart;                  //  Text (  ),
     if FIsSystemHex                 //     SelStart 
       then Text:=IntToHex(FValue,1) //    
       else Text:=IntToStr(FValue);
     SelStart := i;
  except
        if FIsSystemHex
          then Text:=IntToHex(FValue,1)
          else Text:=IntToStr(FValue);
  end;
end;



function TEasyHex.DifferenceValue(v: integer): integer;
begin
  If v=-1 then v:=0;
  If FIsSystemHex
    then result:=Trunc(IntPower(16,v))
    else result:=Trunc(IntPower(10,v));
end;

procedure TEasyHex.KeyDown(var key: Word; shift: TShiftState);
var reverse: integer;
begin
  case key of
      VK_UP: begin
             reverse := pred(Length(Text)-SelStart);
             Inc(FValue,DifferenceValue(reverse));
             if FValue>FValueMax then FValue:=FValueMax;
             if FIsSystemHex
               then Text:=IntToHex(FValue,1)
               else Text:=IntToStr(FValue);
             SelStart := pred(Length(Text)-reverse);
             key:=0;
             end;
      VK_DOWN: begin
               reverse := pred(Length(Text)-SelStart);
               Dec(FValue,DifferenceValue(reverse));
               if FValue<FValueMin then FValue:=FValueMin;
               if FIsSystemHex
                 then Text:=IntToHex(FValue,1)
                 else Text:=IntToStr(FValue);
               SelStart := pred(Length(Text)-reverse);
               key:=0;
               end;
      VK_PRIOR: Value:=FValueMax;
      VK_NEXT: Value:=FValueMin;
      end;
  inherited KeyDown(Key, Shift);
end;



procedure TEasyHex.KeyPress(var key: Char);
begin
  If FIsSystemHex
    then begin                       //  16-
         case key of
             'a','','': key:='A';
             'b','','': key:='B';
             'c','','': key:='C';
             'd','','': key:='D';
             'e','','': key:='E';
             'f','','': key:='F';
             end;
         if not (key in ['0'..'9','A'..'F',#1..#31]) then key:=#0;
         end
    else                             //  10-
         if not (key in ['0'..'9',#1..#31]) then key:=#0;

  If key=#0
    then MessageBeep(0)
    else inherited KeyPress(key);
end;



procedure TEasyHex.Change;
var len, i: Integer;
    bError: Boolean;
    str: string;
begin
  str := Text;
  len := Length(str);
  i := 1;
  bError := false;

  If FIsSystemHex        //  
    then
        While (i<=len) and (len>0) do
             begin
             case str[i] of
                 '0'..'9','A'..'F': ;
                 'a'..'f': str[i] := Chr(Ord(str[i])-32);
                 #1..#32: begin
                          Delete(str,i,1);
                          Dec(i);
                          Dec(len);
                          end;
                 else begin
                      bError := true;
                      Break;
                      end;
                 end;
             Inc(i);
             end
    else
        While (i<=len) and (len>0) do
             begin
             case str[i] of
                 '0'..'9': ;
                 #1..#32: begin
                          Delete(str,i,1);
                          Dec(i);
                          Dec(len);
                          end;
                 else begin
                      bError := true;
                      Break;
                      end;
                 end;
             Inc(i);
             end;


  If bError then begin
                 MessageDlg('    !       .',
                             mtWarning,[mbOk],0);
                 str := LeftStr(str,pred(i));
                 end;
  If str='' then begin
                  FValue := 0;
                  Text := '';
                  EXIT;
                  end;

  try
     If FIsSystemHex
       then
           begin
           FValue := StrToInt('$'+str);
           if FValue<0
             then begin
                  MessageDlg('   ,     Integer.        (2 147 483 647).',
                             mtWarning,[mbOk],0);
                  FValue := 2147483647;
                  end;
           end
       else
           FValue := StrToInt(str);
  except
        MessageDlg('        .      ,     Integer.      (2 147 483 647).'#13#13'    ,     .',
                   mtError,[mbOk],0);
        FValue := 2147483647;
  end;

  If FValue>FValueMax then FValue := FValueMax;

  ValueToText;

  inherited Change;
end;



{   TSpinHex }

constructor TSpinHex.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FButton := TSpinButton.Create(Self);
  FButton.Width := 15;
  FButton.Height := 17;
  FButton.Parent := Self;
  FButton.FocusControl := Self;
  FButton.OnUpClick := UpClick;
  FButton.OnDownClick := DownClick;
end;

destructor TSpinHex.Destroy;
begin
  FButton := nil;
  inherited Destroy;
end;

procedure TSpinHex.UpClick(Sender: TObject);
begin
  If FValue<FValueMax
    then begin
         Inc(FValue);
         ValueToText;
         end;
end;

procedure TSpinHex.DownClick(Sender: TObject);
begin
  If FValue>FValueMin
    then begin
         Dec(FValue);
         ValueToText;
         end;
end;

procedure TSpinHex.CreateWnd;
begin
  inherited CreateWnd;
  SetEditRect;
end;

procedure TSpinHex.SetEditRect;
var
  Loc: TRect;
begin
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc));
  Loc.Bottom := ClientHeight + 1;  {+1 is workaround for windows paint bug}
  Loc.Right := ClientWidth - 15 - 2; // FButton.Width    (15),
  Loc.Top := 0;                      //- ,         
  Loc.Left := 0;
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc));  {debug}
end;

procedure TSpinHex.WMSize(var Message: TWMSize);
var
  MinHeight: Integer;
begin
  inherited;
  MinHeight := GetMinHeight;
    { text edit bug: if size to less than minheight, then edit ctrl does
      not display the text }
  if Height < MinHeight then   
    Height := MinHeight
  else if FButton <> nil then
  begin
    if NewStyleControls and Ctl3D then
      FButton.SetBounds(Width - FButton.Width - 5, 0, FButton.Width, Height - 5)
    else FButton.SetBounds (Width - FButton.Width, 1, FButton.Width, Height - 3);
    SetEditRect;
  end;
end;

function TSpinHex.GetMinHeight: Integer;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  I := SysMetrics.tmHeight;
  if I > Metrics.tmHeight then I := Metrics.tmHeight;
  Result := Metrics.tmHeight + I div 4 + GetSystemMetrics(SM_CYBORDER) * 4 + 2;
end;

procedure TSpinHex.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
{  Params.Style := Params.Style and not WS_BORDER;  }
  Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN;
end;

procedure TSpinHex.CMEnter(var Message: TCMGotFocus);
begin
  if AutoSelect then SelectAll;
  inherited;
end;

end.
